# -*- coding: utf-8 -*-
"""
Package parameters:
-----------------------
 Flask-Mercury base parameter type definition.
"""
import inspect


class MetaParameter(type):
    """
    Mercury Parameter metaclass.
    """
    def __instancecheck__(cls, instance):
        """
        The magic function called to implement isinstance(..) functionality.
        :param instance: the object being evaluated.
        :return: true if instance should be considered a (direct or indirect) instance of class.
        """
        return cls.__instancecheck__(instance)


class BaseParameter(object, metaclass=MetaParameter):
    """
    Base Mercury parameter class.
    """
    flask_type = ""

    @classmethod
    def to_swagger(cls, param, doc):
        """
        Base function parameter spec serialization function.
        """
        spec = {
            "name": param.name,
        }
        if doc is not None:
            spec["description"] = doc

        if param.default is not inspect._empty:
            spec["required"] = False
        else:
            spec["required"] = True

        return spec

    @classmethod
    def __instancecheck__(cls, instance):
        """
        The magic function called to implement isinstance(..) functionality.
        :param instance: the object being evaluated.
        :return: true if instance should be considered a (direct or indirect) instance of class.
        """
        pass




# -*- coding: utf-8 -*-
"""
Module flask_resource.py
---------------------
    The base flask Api resource implementation.
    Provides the base end point functionality such as route register request auto serialization
    and method dispatch.
"""
from ..meta_resource import BaseResource
from mercury.parameters.path import BasePathParameter
from flask import request, jsonify
from mercury.api_model import ApiModel
from flask import Flask
import inspect
import json
from mercury import exceptions
from io import StringIO
from werkzeug import exceptions as http_exceptions
from mercury.resource import _verbs
from . import path_utils
from flask.globals import current_app


class FlaskResource(BaseResource):
    """
    A Base API endpoint class implementation.
    """
    ROUTE = None
    __model__ = None

    def __init__(self):
        """
        Resource initialization function.
        """
        super().__init__()
        # request object shortcut
        self.request = request

    def dispatch_request(self, *args, **kwargs):
        """
        Parses the body of an post request into a Mapping definition object before
        dispatching the request to the endpoint implementation.
        """
        # identify the http request verb
        meth = getattr(self, request.method.lower(), None)
        if meth is None and request.method == 'HEAD':
            meth = getattr(self, 'get', None)
        # sanity check on verb implementation
        if meth is None:
            raise http_exceptions.MethodNotAllowed(
                None,
                "The api endpoint {}, does not support the {} http/https verb.".format(
                    request.path, request.method
                )
            )

        args_ = list()
        kwargs_ = dict()
        model = None
        for param in meth.__params__:
            if issubclass(param.annotation, (ApiModel,)):
                model = param.annotation
            if param.name != "self":
                if param.default is inspect._empty:
                    default = None
                else:
                    default = param.default
                p = param.annotation.from_request(param.name, meth, default)
                if default is None:
                    args_.append(p)
                else:
                    kwargs_[param.name] = p

        if model is None and len(request.data) > 0:
            raise http_exceptions.InternalServerError(
                "There is an error on the resource definition. "
                "Could not parse the request body."
            )
        response = meth(*args_, **kwargs_)
        # response = super().dispatch_request(*args_, **kwargs_)
        if isinstance(response, (ApiModel,)):
            response = response.to_dict()

        response = json.dumps(response, allow_nan=False, ensure_ascii=False)
        response = current_app.response_class(
            response,
            mimetype=current_app.config['JSONIFY_MIMETYPE']
        )
        return response

    @classmethod
    def register(cls, app:Flask, api_title):
        """
        Register the resource endpoint route and initialize its path parameters.
        :param app: a reference to the 'flask.Flask' app.
        :param api_title: the api title.
        """
        base_path = cls.get_base_path()
        # get the resource base path
        # checks for route meta data
        base_parameters = path_utils.get_route_parameters(base_path)

        # for each http verb performs a sanity check on the path parameters
        for verb in _verbs:
            if hasattr(cls, verb):
                impl = getattr(cls, verb)
                # checks path parameters definition
                cls.check_base_param_equality(base_parameters, impl)

                # builds final paths
                # for each additional parameter
                # appends to the path
                cls.add_url_rules(app, impl, verb, api_title)

    @classmethod
    def get_base_path(cls):
        """
        Returns the resource base route path.
        If no cls.ROUTE was provided it uses the class dotted path to build a base route for the resource.
        :return: the resource base route path
        """
        base_path = cls.ROUTE
        if base_path is None:
            # otherwise uses the module dotted path as path
            mod = cls.__module__.split('.')
            # remove file name from path
            if len(mod) > 1:
                mod.pop()
            if cls.__name__.lower() != mod[-1]:
                mod = "/".join(mod)
                base_path = "/{}/{}".format(mod, cls.__name__.lower())
            else:
                mod = "/".join(mod)
                base_path = "/{}".format(mod)

        return base_path

    @classmethod
    def add_url_rules(cls, app, impl, verb_name, api_title):
        """
        Builds a complete route path and adds the route rule to the impl verb.
        :param app: A flask application object;
        :param impl: the verb implementation.
        :param verb_name: the verb name e.g. get, post, put .....
        :param api_title: the api title.
        """
        base_path = cls.get_base_path()
        # builds final paths
        # for each additional parameter
        # appends to the path
        str_builder = StringIO()
        first_default_flag = False
        # base_path must start with the api_title
        if not base_path.startswith(api_title):
            str_builder.write("/" + api_title)
        str_builder.write(base_path)
        for parameter in impl.__params__:
            if parameter.default is not inspect._empty and not first_default_flag:
                app.add_url_rule(
                    str_builder.getvalue(),
                    view_func=cls.as_view("{}_{}".format(verb_name,cls.__name__)),
                    methods=[verb_name.upper(), ]
                )
                first_default_flag = True

            # adds path_param to the base_path
            if issubclass(parameter.annotation, BasePathParameter):
                if parameter.name not in base_path:
                    path_param = "/<{}:{}>".format(parameter.annotation.flask_type, parameter.name)
                    str_builder.write(path_param)

        view_name = "{}_{}_full".format(verb_name, cls.__name__)
        app.add_url_rule(
            str_builder.getvalue(),
            view_func=cls.as_view(view_name),
            methods=[verb_name.upper(), ]
        )

    @staticmethod
    def check_base_param_equality(base_params, impl):
        """
        checks whether all items in both list of parameters are equals.
        :param base_params: list of base path parameters
        :param impl: a reference to a verb method implementation.
        :return: true if all items in the sequence of both lists are equals.
        :raises: mercury.exceptions.ResourceSpecError
        """
        if base_params is not None:
            # removes args, kwargs and self from the list of impl_params
            params = [
                p.name for p in impl.__params__
                if p.name not in ["args", "kwargs", "self"]
                and not issubclass(p.annotation, (ApiModel,))
            ]
            if len(params) == 0:
                return True
            # check list equality
            for i in range(len(base_params)):
                if params[i] != base_params[i]:
                    # if some divergence is found,
                    # then returns false
                    raise exceptions.ResourceSpecError(
                        "Wrong method parameter definition on '{impl}'. "
                        "The all base_path parameters must be defined in the method signature,"
                        " in the ordering that they appear on the base path. "
                        "Additional parameters may be declared after the ones defined on the base_path."
                        "For more information please refers to the mercury documentation."
                        "Method name: {impl}."
                        "Method declared parameters: {meth_params}."
                        "Base path parameters: {base_params}".format(
                            impl=impl.__qualname__,
                            meth_params=params,
                            base_params=base_params
                        )
                    )
        # return true if every thing is ok
        return True


# -*- coding: utf-8 -*-
"""
Module utils.py
-----------------
 A set of utility functions that are used to build resources and swagger specification.
"""
import re
import sys
from simple_mappers.object_mapper import JsonObject
from importlib import import_module
import six


def trim(docstring):
    """
    trim function from PEP-257
    
    | **function adapted from**
    |    https://github.com/openstack/rally/blob/7153e0cbc5b0e6433313a3bc6051b2c0775d3804/rally/common/plugin/info.py
    
    """
    if not docstring:
        return ""
    # Convert tabs to spaces (following the normal Python rules)
    # and split into a list of lines:
    lines = docstring.expandtabs().splitlines()
    # Determine minimum indentation (first line doesn't count):
    indent = sys.maxsize
    for line in lines[1:]:
        stripped = line.lstrip()
        if stripped:
            indent = min(indent, len(line) - len(stripped))
    # Remove indentation (first line is special):
    trimmed = [lines[0].strip()]
    if indent < sys.maxsize:
        for line in lines[1:]:
            trimmed.append(line[indent:].rstrip())
    # Strip off trailing and leading blank lines:
    while trimmed and not trimmed[-1]:
        trimmed.pop()
    while trimmed and not trimmed[0]:
        trimmed.pop(0)

    # Current code/unittests expects a line return at
    # end of multiline docstrings
    # workaround expected behavior from unittests
    if "\n" in docstring:
        trimmed.append("")

    # Return a single string:
    return "\n".join(trimmed)


def reindent(string):
    """
    | **function adapted from**
    |    https://github.com/openstack/rally/blob/7153e0cbc5b0e6433313a3bc6051b2c0775d3804/rally/common/plugin/info.py
    
    """
    return "\n".join(l.strip() for l in string.strip().split("\n"))


def import_string(dotted_path):
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.
    """
    if '.' not in dotted_path:
        return import_module(dotted_path)
    else:

        try:
            module_path, class_name = dotted_path.rsplit('.', 1)
        except ValueError:
            msg = "%s doesn't look like a module path" % dotted_path
            six.reraise(ImportError, ImportError(msg), sys.exc_info()[2])

        module = import_module(module_path)

        try:
            return getattr(module, class_name)
        except AttributeError:
            msg = 'Module "{}" does not define a "{}" attribute/class'.format(
                dotted_path, class_name
            )
            six.reraise(ImportError, ImportError(msg), sys.exc_info()[2])


def is_integer(value):
    try:
        value = int(value)
        return True
    except:
        return False


# -*- coding: utf-8 -*-
"""
Package query:
-----------------------
 Flask-Mercury query parameter type definitions.
"""
from .base import BaseQueryParameter
from .boolean import Boolean
from .float_param import Float
from .integer import Integer
from .string import String, Any
from .uuid import Uuid



# -*- coding: utf-8 -*-
"""
Module resource.py
---------------------
    The base Api swagger resource implementation.
    Provides the base functionality to serialize the resource meta information
    to the swagger spec.
"""
from mercury.resource import BaseResource
from mercury.api_model import ApiModel
import inspect
from simple_mappers.object_mapper import JsonObject
from mercury import utils
from mercury.resource import _verbs
from . import path_utils
from mercury.parameters.base import BaseParameter
from mercury.parameters.path import BasePathParameter
from mercury.parameters.query import BaseQueryParameter


class SwaggerResource(BaseResource):
    """
    A Base Swagger resource that adds metadata serialization to the swagger spec format.
    """
    ROUTE = None
    TAGS = None

    @classmethod
    def to_swagger(cls):
        """
        Swagger spec serialization function.
        :param cls: The class to be serialized.
        """
        # each verb builds a swagger path list
        resource_spec = JsonObject()
        resource_spec.paths = dict()
        resource_spec.tags = [{"name": tag} for tag in cls.get_tags()]
        for verb in _verbs:
            verb_specs = cls.get_verb_spec(verb)
            # now adds the verb to the resource spec
            for path, verb_spec in verb_specs.items():
                if path in resource_spec.paths:
                    resource_spec.paths[path][verb] = verb_spec[verb]
                else:
                    resource_spec.paths[path] = {
                        verb: verb_spec[verb]
                    }
        return resource_spec

    @classmethod
    def get_tags(cls):
        """
        Returns the resource tags.
        :param cls: the resource class being serialized to the swagger spec json.
        """
        # checks for Meta data
        if cls.TAGS is not None:
            # checks whether a tag meta data is provided
            return cls.TAGS
        else:
            # if no tag was provided then it will use the resource root __module__
            # path as tag
            tags = cls.__module__.split('.')
            if len(tags) > 1:
                return [tags[1]]
            else:
                return [tags[0]]

    @classmethod
    def get_verb_spec(cls, verb):
        """
        Builds and returns the api spec for the resource verb.
        :return: The swagger Resource verb specification.
        """
        paths_spec = dict()
        paths = cls.get_paths(verb)
        # for each path in the paths list
        #  we build the swagger paths part of the schema
        for path, impl in paths:  # impl: reference to the resource method that implements the verb
            spec = cls.build_verb_spec(path, impl)
            # now adds the verb to path spec
            if path in paths_spec:
                paths_spec[path][verb] = spec
            else:
                paths_spec[path] = {
                    verb: spec
                }
        return paths_spec

    @classmethod
    def get_paths(cls, verb):
        """
        returns a list of paths for the verb received as parameter.
        :param verb: a verb name. Allowed values are "get", "post", "put", "delete", "head", "patch".
        :return: a list of tuples (path, impl),
        where *path* is the string path to the endpoint
        and *impl* is a reference for the class resource method that implements the endpoint.
        """
        # initializes the paths list
        paths = list()
        base_path = path_utils.format_base_path_swagger(cls)
        if hasattr(cls, verb):
            # sanity checks whether all base path parameters are defined on the resource function
            path_utils.check_route_and_verb_path_parameter_specifications(cls, verb)

            method = getattr(cls, verb, None)

            additional_params = path_utils.find_additional_params(cls, method)
            # no additional parameter was found
            if len(additional_params) == 0:
                paths = [(base_path, method)]
            else:  # verb defines more parameters than the base path
                required_params = path_utils.filter_required_params(additional_params)
                if len(required_params) > 0:
                    path = "{}/{}".format(
                        base_path,
                        "/".join(required_params)
                    )
                    # appends the parameter to the path list
                    paths.append((path, method))

                not_required_params = path_utils.filter_not_required_params(additional_params)
                if len(not_required_params) > 0:
                    # then appends one additional path to the paths list
                    # with the paths definition
                    path = "{}/{}".format(
                        base_path,
                        "/".join(not_required_params)
                    )
                    paths.append((path, method))

        return paths

    @classmethod
    def build_verb_spec(cls, path, verb):
        """
        Builds the swagger spec for the pair pair (path, verb)
        :param path: a resource endpoint path
        :param verb: a function that implements the verb.
        :return: the verb swagger specification.
        """
        verb_spec = JsonObject()
        verb_spec.tags = cls.get_tags()
        verb_spec.summary = verb.__pdoc__.summary
        verb_spec.description = verb.__pdoc__.description
        verb_spec.consumes = verb.__pdoc__.consumes
        verb_spec.produces = verb.__pdoc__.produces
        verb_spec.parameters = cls.get_verb_parameters_spec(path, verb)

        # finally builds the responses specification
        verb_spec.responses = cls.get_verb_responses_spec(verb)
        return verb_spec

    @classmethod
    def get_verb_parameters_spec(cls, path, verb):
        """
        Returns the swagger parameters spec for the verb implementation and path pair.
        :param path: a resource path.
        :param verb: a verb implementation.
        :return: swagger parameters specification.
        """
        spec = list()
        # foreach parameter defined in the verb implementation
        for par in verb.__params__:
            # path_param = '{'+par.name+'}'
            # if path_param in path:
            if issubclass(par.annotation, BasePathParameter):
                param_spec = par.annotation.to_swagger(
                    par,
                    verb.__pdoc__.params.get(par.name)
                )
                spec.append(param_spec)

            elif issubclass(par.annotation, ApiModel):
                if "multipart/form-data" in verb.__pdoc__.consumes:
                    parameter_set = par.annotation.to_swagger_as_form()
                    spec.extend(parameter_set)
                else:
                    param_spec = par.annotation.to_swagger_as_parameter(
                        par,
                        verb.__pdoc__.params.get(par.name)
                    )
                    spec.append(param_spec)

            elif issubclass(par.annotation, BaseQueryParameter):
                param_spec = par.annotation.to_swagger(
                    par,
                    verb.__pdoc__.params.get(par.name)
                )
                spec.append(param_spec)

        return spec

    @classmethod
    def get_verb_responses_spec(cls, verb):
        """
        Returns the verb swagger responses spec for the given verb.
        :param verb: a verb implementation.
        :return: the verb swagger reponses specification.
        """
        spec = dict()
        # if it was declared the return type
        if issubclass(verb.__returns__, ApiModel):
            spec["200"] = verb.__returns__.to_swagger_as_response(verb.__pdoc__.returns)
        else:
            spec["200"] = {
                "description": verb.__pdoc__.returns,
                # 'type': self.TYPE_MAP.get(impl.__returns__.__name__)
            }

        for ex, description in verb.__pdoc__.raises.items():
            # ex is a string containing or the error code
            # or the dotted path to the exception
            if not utils.is_integer(ex):  # is it is a dotted path
                _ex = utils.import_string(ex)
                # sanity check exception type
                if not hasattr(_ex, "code"):
                    raise TypeError(
                        "The exception {} specified on the doc sting is no an http exception"
                        "and does not defines a error code."
                        "Exceptions specified on the docstring must by an http exception.".format(ex)
                    )
                spec[str(_ex.code)] = {"description": description}
            else:  # otherwise it is just a error code
                spec[ex] = {"description": description}

        return spec


# -*- coding: utf-8 -*-
"""
Date property module - defines date attributes.
"""
from simple_mappers import properties
from .base import BaseProperty
from werkzeug import exceptions


class DateProperty(properties.DateProperty, BaseProperty):
    """
    Stores a date
    """

    def inflate(self, value):
        """Returns the value that should be used to fill the object being mapped."""
        try:
            value = super().inflate(value)
            return value
        except TypeError as e:
            raise exceptions.BadRequest(str(e))
        except ValueError as e:
            name = self.mapping_from or self.property_name
            raise exceptions.BadRequest(
                "On Property: {}\n"
                "Value: {}\n"
                "Error: Date property '{}' does not match format '{}'.".format(name, value, name, self.mask)
            )

    def to_swagger(self, schema:dict):
        """
        Serializes attribute to swagger specification format.
        :param schema: a swagger spec dict.
        :return: the modified swagger spec.
        """
        super(DateProperty, self).to_swagger(schema)
        schema["type"] = "string"
        schema["format"] = self.mask
        return schema


# -*- coding: utf-8 -*-
"""
Module base_tests.py
----------------------
A test suite for the base path parameter.
"""
import unittest
from mercury.parameters.path.base import BasePathParameter
from mercury.doc_parser import DocString
import inspect


def test_function(parameter:BasePathParameter="test"):
    """
    A function that defines a base parameter with default value.
    :param parameter: the function parameter with default value.
    """
    pass


class BaseParameterTests(unittest.TestCase):
    """
    A test suite for the base parameter class.
    """

    def test_should_serialize_spec_parameter_required_correctly(self):
        """
        Tests if the parameter spec is serialized correctly.
        """

        signature = inspect.signature(test_function)
        method_params = list(signature.parameters.values())
        doc_parser = DocString()
        doc_str = doc_parser.parse(test_function.__doc__)
        par = method_params[0]
        spec = BasePathParameter.to_swagger(par, doc_str.to_dict().get("params").get("parameter"))

        self.assertEqual(spec['in'], "path")


if __name__ == "__main__":
    unittest.main()


# -*- coding: utf-8 -*-
"""
Module base.py
----------------- 
    Defines a base class property.
    Properties are used to give meta-data to class attributes.
"""
from simple_mappers import properties
from simple_mappers.properties import mapping_error
from werkzeug import exceptions


class BaseProperty(properties.BaseProperty):
    """
    Base class for object properties
    """

    def __init__(
            self, required=False, default=None,
            mapping_name=None, cascade=True, validate_func=None,
            description=None, example=None, **kwargs
    ):
        """
        A Base class property definition. 
        :param required: True if the attribute is required.
                         When true and it is not possible to map the attribute from the other object,
                         it raises a value error exception. 
        :param default: default value or function generator.
        :param mapping_name: the name of the attribute in the other object
        :param cascade: when true mappers try to map the the property from the other object properties.
        :param validate_func: a function used to validate attribute values.
        :param description: a property description sting.
        :param example: a value example for the attribute.
        :param kwargs:
        """
        super().__init__(
            required=required,
            default=default,
            mapping_name=mapping_name,
            cascade=cascade,
            validade_func=validate_func,
            **kwargs
        )

        self.description = None
        if description is not None:
            self.description = description
        self.example = example

    def inflate(self, value):
        """
        returns default or value. 
        """
        # applies the validation function
        try:
            return super().inflate(value)
        except mapping_error.InvalidAttributeValueError as e:
            raise exceptions.BadRequest(str(e))
        except mapping_error.RequiredPropertyError as e:
            raise exceptions.BadRequest(str(e))

    def to_swagger(self, schema: dict):
        """
        Serializes attribute to swagger specification format.
        :param schema: a swagger spec dict.
        :return: the modified swagger spec.
        """
        if self.required:
            if "required" not in schema:
                schema["required"] = list()
            name = self.property_name if not self.mapping_from else self.mapping_from
            schema["required"].append(name)
        if self.default is not None:
            schema["default"] = self.default

        if self.description is not None:
            schema["description"] = self.description
        return schema


class ChoiceProperty(properties.ChoiceProperty, BaseProperty):
    """
    Defines a base choosable base property.
    """

    def inflate(self, value):
        """Returns the value that should be used to fill the object being mapped."""
        try:
            value = super().inflate(value)
            return value
        except mapping_error.InvalidAttributeValueError as e:
            raise exceptions.BadRequest(str(e))

    def deflate(self, value, **kwargs):
        """Returns the value that should be used to send to the object being mapped."""
        try:
            value = super().deflate(value, **kwargs)
            return value
        except ValueError as e:
            raise exceptions.BadRequest(str(e))

    def to_swagger(self, schema:dict):
        """
        Serializes attribute to swagger specification format.
        :param schema: a swagger spec dict.
        :return: the modified swagger spec.
        """
        super(ChoiceProperty, self).to_swagger(schema)
        schema["type"] = "string"
        if self.choices is not None:
            schema["enum"] = self.choices
        return schema


class NormalProperty(properties.NormalProperty, BaseProperty):
    """
    Base class for normalized properties.
    """
    def to_swagger(self, schema:dict):
        """
        Serializes attribute to swagger specification format.
        :param schema: a swagger spec dict.
        :return: the modified swagger spec.
        """
        super(NormalProperty, self).to_swagger(schema)
        schema["type"] = "string"
        return schema



# -*- coding: utf-8 -*-
"""
Package parameters:
-----------------------
 Flask-Mercury base parameter type definition.
"""
from werkzeug import exceptions
from werkzeug.datastructures import FileStorage
from .base import BaseProperty


class FileProperty(BaseProperty):
    """
    A File parameter definition.
    """
    flask_type = "file"

    # def __new__(cls, name, value):
    #     """
    #     A file parameter constructor.
    #     :param value: a value to be casted to int.
    #     :return: the value casted to int type
    #     """
    #
    #     if isinstance(value, io.IOBase):
    #         return value
    #     else:
    #         raise BadRequest(
    #             "Bad file received '{}'. "
    #             "We could not get the file stream, instead we got '{}'."
    #             "".format(name, value)
    #         )

    def inflate(self, value):
        value = super().inflate(value)
        if value is not None:

            if not isinstance(value, FileStorage):
                name = self.mapping_from or self.property_name
                raise exceptions.BadRequest(
                    "On property: '{}'.\n"
                    "Value: '{}'.\n"
                    "Error: A file was expected.".format(name, value)
                )

        return value

    def to_swagger(self, schema):
        """
        Swagger spec path parameter serialization function.
        :param param: function parameter pointer.
        :return: the parameter spec.
        """
        spec = super().to_swagger(schema)
        spec['type'] = 'file'
        spec['format'] = 'binary'
        spec['in'] = 'formData'

        return spec



"""
integer property module - defines integer attributes.
"""
from simple_mappers import properties
from .base import BaseProperty
from werkzeug import exceptions


class IntegerProperty(properties.IntegerProperty, BaseProperty):
    """
    Stores an Integer value
    """

    def inflate(self, value):
        """Returns the value that should be used to fill the object being mapped."""
        try:
            value = super().inflate(value)
            return value
        except ValueError:
            name = self.mapping_from or self.property_name
            raise exceptions.BadRequest(
                "On property: '{}'.\n"
                "Value: '{}'.\n"
                "Error: Expected int type.".format(name, value)
            )

    def to_swagger(self, schema: dict):
        """
        Serializes attribute to swagger specification format.
        :param schema: a swagger spec dict.
        :return: the modified swagger spec.
        """
        super(IntegerProperty, self).to_swagger(schema)
        schema["type"] = "integer"
        return schema



# -*- coding: utf-8 -*-
"""
Package parameters:
-----------------------
 Flask-Mercury *query* parameter type definitions.
"""
import inspect
from ..base import BaseParameter
from flask import request


class BaseQueryParameter(BaseParameter):
    """
    Base path parameter class
    """
    @classmethod
    def to_swagger(cls, param, doc):
        """
        Base query parameter spec serialization function.
        """
        spec = super().to_swagger(param, doc)
        spec["in"] = "query"

        return spec

    @classmethod
    def from_request(cls, name, meth, default=None):
        value = request.args.get(name, default)
        if value is not None:
            return cls(name, value)
        else:
            return None


# -*- coding: utf-8 -*-
"""
Module array_property.py
----------------------------
    Defines array attributes.
"""
from simple_mappers import properties
import collections
from .base import BaseProperty
from .object_property import ObjectProperty
from werkzeug import exceptions
from ..api_model import ApiModel


class ArrayProperty(properties.ArrayProperty, BaseProperty):
    """
    Stores a list of items.
    """
    def inflate(self, value):
        """Returns the value that should be used to fill the object being mapped."""
        try:
            return super().inflate(value)
        except TypeError as e:
            raise exceptions.BadRequest(str(e))

    def deflate(self, value, **kwargs):
        """Returns the value that should be used to send to the object being mapped."""
        try:
            return super().deflate(value, **kwargs)
        except TypeError as e:
            raise exceptions.BadRequest(str(e))

    def to_swagger(self, schema:dict):
        """
        Swagger serialization function.
        :param schema: a dictionary used to represent and store the schema.
        """
        super(ArrayProperty, self).to_swagger(schema)
        schema["type"] = "array"
        if self.itens_type is not None:
            if isinstance(self.itens_type, type) and issubclass(self.itens_type, (ApiModel,)):
                schema["items"] = {
                    "$ref": "#/definitions/{}".format(self.itens_type.__name__)
                }
            else:
                schema["items"] = self.itens_type.to_swagger(dict())
        return schema


# -*- coding: utf-8 -*-
"""
Module api_model.py 
---------------------
    Defines the base class to the resource body.
"""
from simple_mappers import MapDefinition
import inspect
from flask import request
import json
from werkzeug import exceptions as http_exceptions


class ApiModel(MapDefinition):
    """An base ApiModel.
    ApiModel defines the structure of the resources body message.
    """
    def __init__(self, map_undefined=False, **kwargs):
        super().__init__(map_undefined, **kwargs)

    @classmethod
    def to_swagger(cls, spec):
        """
        A Model class to swagger schema serialization.
        :param spec: A dict object used to store and represent the swagger specification.
        """
        model_schema = dict()

        model_schema["type"] = "object"
        model_schema["properties"] = dict()
        for property_name, property_def in cls.__properties__.items():
            p_name = property_name if property_def.mapping_from is None else property_def.mapping_from
            model_schema["properties"][p_name] = property_def.to_swagger(dict())
        spec.definitions[cls.__name__] = model_schema
        return spec

    @classmethod
    def to_swagger_as_parameter(cls, param, doc):
        """
        Serializes a api model as a resource verb parameter.
        :param param: a verb implementation parameter.
        :param doc: a parameter doc string.
        :return: the swagger parameter specification.
        """
        # parameter spec
        pspec = dict()
        pspec['name'] = param.name
        pspec['description'] = doc
        if param.default is inspect._empty:
            pspec["required"] = True
        else:
            pspec["required"] = False

        pspec["in"] = "body"
        pspec["schema"] = {"$ref": "#/definitions/{}".format(cls.__name__)}
        return pspec

    @classmethod
    def to_swagger_as_form(cls):
        """
        Serializes a api model as a set of form's parameter.
        :param param: a verb implementation parameter.
        :param doc: a parameter doc string.
        :return: the swagger parameter specification.
        """
        parameters_set = list()
        for property_name, property_def in cls.__properties__.items():
            pspec = dict()
            p_name = property_name if property_def.mapping_from is None else property_def.mapping_from
            pspec['name'] = p_name
            pspec = property_def.to_swagger(pspec)
            pspec["in"] = "formData"
            parameters_set.append(pspec)

        return parameters_set

    @classmethod
    def to_swagger_as_response(cls, doc):
        spec = {
            "schema": {
                "$ref": "#/definitions/{}".format(cls.__name__)
            }
        }
        if doc is not None:
            spec["description"] = doc
        return spec

    @classmethod
    def from_request(cls, name, meth, default=None):
        if "multipart/form-data" in meth.__pdoc__.consumes:
            data = dict()
            data.update({key: value for key, value in request.form.items()})
            data.update({key: value for key, value in request.files.items()})

            try:
                if len(data) > 0:
                    return cls.from_dict(data)
                else:
                    return default
            except Exception as e:
                raise http_exceptions.BadRequest(str(e))

        else:
            if request.data is not None:
                # decode the request data body
                data = request.data.decode("utf-8").strip()
                try:
                    if data != "":
                        data = json.loads(data)
                        return cls.from_dict(data)
                    else:
                        return default
                except Exception as e:
                    raise http_exceptions.BadRequest(str(e))



# -*- coding: utf-8 -*-
"""
Module utils.py
-----------------
 A set of utility functions that are used to build resources and swagger specification.
"""
import pyparsing as pp
from simple_mappers.object_mapper import JsonObject


param_mark = pp.CaselessLiteral(":param").suppress()
return_mark = pp.Or([
    pp.CaselessLiteral(":return"),
    pp.CaselessLiteral(":returns")
]).suppress()

raise_mark = pp.Or([
    pp.CaselessLiteral(":raise"),
    pp.CaselessLiteral(":raises")
]).suppress()

consumes_mark = pp.CaselessLiteral(":consumes:")
consumes_mark = consumes_mark.suppress()
produces_mark = pp.CaselessLiteral(":produces:")
produces_mark = produces_mark.suppress()
version_mark = pp.CaselessLiteral(":version:")
version_mark = version_mark.suppress()

other_marks = pp.Or(
    [
        pp.CaselessLiteral(":arg"),
        pp.CaselessLiteral(":argument"),
        pp.CaselessLiteral(":cvar"),
        pp.CaselessLiteral(":ivar"),
        pp.CaselessLiteral(":except"),
        pp.CaselessLiteral(":exception"),
        pp.CaselessLiteral(":key"),
        pp.CaselessLiteral(":keyword"),
        pp.CaselessLiteral(":parameter"),
        pp.CaselessLiteral(":rtype"),
        pp.CaselessLiteral(":type"),
        pp.CaselessLiteral(":var"),
    ]
)

any_start_mark = pp.Or([
    param_mark,
    return_mark,
    raise_mark,
    consumes_mark,
    produces_mark,
    version_mark,
    other_marks
]).suppress()


class BaseElement(pp.ParseElementEnhance):
    """
    BaseGrammar element class.
    """
    def __init__(self, savelist=False, name=None):
        exprs = self.get_exprs()
        # super class init call.
        super().__init__(exprs, savelist)

        if name is None:
            self.setName(self.__class__.__name__.lower())
            self.setResultsName(self.__class__.__name__.lower())
        else:
            self.setName(name)
            self.setResultsName(name)


    def get_exprs(self):
        """
        Py parsing expression definition.
        :return: an expression
        """
        raise NotImplementedError("Abstract method.")

    def setName(self, name):
        """
        Define name for this expression, makes debugging and exception messages clearer.

        Example::
            Word(nums).parseString("ABC")  # -> Exception: Expected W:(0123...) (at char 0), (line:1, col:1)
            Word(nums).setName("integer").parseString("ABC")  # -> Exception: Expected integer (at char 0), (line:1, col:1)
        :param name: the name to be set.
        :return: self
        """
        self.expr = self.expr.setName(name)

        return super().setName(name)

    def setResultsName(self, name, listAllMatches=False):
        """
        Define name for referencing matching tokens as a nested attribute
        of the returned parse results.
        NOTE: this returns a *copy* of the original C{ParserElement} object;
        this is so that the client can define a basic element, such as an
        integer, and reference it in multiple places with different names.

        You can also set results names using the abbreviated syntax,
        C{expr("name")} in place of C{expr.setResultsName("name")} -
        see L{I{__call__}<__call__>}.

        Example::
            date_str = (integer.setResultsName("year") + '/'
                        + integer.setResultsName("month") + '/'
                        + integer.setResultsName("day"))

            # equivalent form:
            date_str = integer("year") + '/' + integer("month") + '/' + integer("day")
        :param name: the name to be set
        :param listAllMatches:
        :return:
        """
        self.expr = self.expr.setResultsName(name, listAllMatches)
        return super().setResultsName(name, listAllMatches)


class Param(BaseElement):
    """
    Represents a parameter in doc string.
    """
    def get_exprs(self):

        param_name = pp.Word(pp.alphanums + "_").setName("name").setResultsName("name")

        end_mark = pp.Or([
            any_start_mark.suppress(),
            pp.stringEnd
        ]).suppress()
        param_description = pp.SkipTo(end_mark) \
            .setName("description").setResultsName("description")

        return pp.Group(
            param_mark +
            param_name +
            pp.CaselessLiteral(":").suppress() +
            param_description +
            pp.FollowedBy(end_mark).suppress()
        ).setName("param").setResultsName("param")


class Raise(BaseElement):
    """
    Represents a raise doc string.
    """

    def get_exprs(self):
        error = pp.Word(pp.alphanums + "_.").setName("error").setResultsName("error")
        # error_description = pp.Word(pp.printables + " ").setName("description").setResultsName("description")
        end_mark = pp.Or([
            any_start_mark.suppress(),
            pp.stringEnd
        ]).suppress()
        error_description = pp.SkipTo(end_mark) \
            .setName("description").setResultsName("description")
        return pp.Group(
            raise_mark +
            error +
            pp.CaselessLiteral(":").suppress() +
            error_description +
            pp.FollowedBy(end_mark).suppress()
        ).setName("raise").setResultsName("raise")


class Return(BaseElement):
    """
    Represents a return/returns in doc string.
    """

    def get_exprs(self):
        end_mark = pp.Or([
            any_start_mark.suppress(),
            pp.stringEnd
        ]).suppress()
        description = pp.SkipTo(end_mark) \
            .setName("description").setResultsName("description")
        return pp.Group(
            return_mark +
            description +
            pp.FollowedBy(end_mark).suppress()
        ).setName("raise").setResultsName("raise")


class Consumes(BaseElement):
    """
    Represents a consumes in doc string.
    """

    def get_exprs(self):
        description = pp.Word(pp.printables + " ").setName("description").setResultsName("description")
        end_mark = pp.Or([
            any_start_mark.suppress(),
            pp.stringEnd
        ]).suppress()
        return pp.Group(
            consumes_mark +
            description +
            pp.FollowedBy(end_mark).suppress()
        ).setName("consumes").setResultsName("consumes")


class Produces(BaseElement):
    """
    Represents a consumes in doc string.
    """

    def get_exprs(self):
        description = pp.Word(pp.printables + " ").setName("description").setResultsName("description")
        end_mark = pp.Or([
            any_start_mark.suppress(),
            pp.stringEnd
        ]).suppress()
        return pp.Group(
            produces_mark +
            description +
            pp.FollowedBy(end_mark).suppress()
        ).setName("produces").setResultsName("produces")


class Version(BaseElement):
    """
    Represents a consumes in doc string.
    """

    def get_exprs(self):
        description = pp.Word(pp.printables + " ").setName("description").setResultsName("description")
        end_mark = pp.Or([
            any_start_mark.suppress(),
            pp.stringEnd
        ]).suppress()
        return pp.Group(
            version_mark +
            description +
            pp.FollowedBy(end_mark).suppress()
        ).setName("version").setResultsName("version")


class Other(BaseElement):
    """
    Any other possible mark not used to build swagger documentation.
    """

    def get_exprs(self):

        end_mark = pp.Or([
            any_start_mark.suppress(),
            pp.stringEnd
        ]).suppress()
        body = pp.SkipTo(end_mark) \
            .setName("description").setResultsName("description")

        return pp.Group(
            other_marks +
            body +
            pp.FollowedBy(end_mark).suppress()
        ).setName("other").setResultsName("other")



class DocString(BaseElement):
    """
    Represents a python doc string parser element.
    """
    def get_exprs(self):

        end_mark = pp.Or([
            any_start_mark.suppress(),
            pp.stringEnd
        ]).suppress()

        description = pp.SkipTo(end_mark)\
            .setName("description").setResultsName("description")

        any_doc = pp.Or([
            Param().setName("param").setResultsName("param"),
            Return().setName("return").setResultsName("return"),
            Raise().setName("raise").setResultsName("raise"),
            Consumes().setResultsName("consumes").setName("consumes"),
            Produces().setName("produces").setResultsName("produces"),
            Version().setResultsName("version").setResultsName("version"),
            Other().setResultsName("other").setResultsName("other").suppress(),
        ])
        # pp.locatedExpr()
        docString = description + \
                    pp.FollowedBy(end_mark).suppress() + \
                    pp.ZeroOrMore(any_doc).setName("attrs").setResultsName("attrs")

        return docString.setName("doc").setResultsName("doc")

    def parse(self, docstring):
        """
        Parses the given docstring.
        :param docstring: a docstring to be parsed.
        :type docstring: str
        :return: a SimpleMappers.JsonObject of form
            `
               {
                   "summary": ..., \n
                   "description": ..., \n
                   "params": {"name": "doc"}, \n
                   "returns": ..., \n
                   "consumes":...., \n
                   "produces":....,\n
                   "raises":....., \n
               }
            `
        """
        # result variables declaration
        params = dict()
        returns = None
        raises = dict()
        consumes = list()
        produces = list()
        version = None
        short_description = ""
        description = ""

        # split out the first row from the the docString
        # first row is used as short description.
        if docstring is not None:
            doc_rows = docstring.split("\n", 1)
            if len(doc_rows) > 1:
                short_description = doc_rows[0].strip(' ')
                parsed_doc = self.parseString(docstring)
            else:
                short_description = ""
                parsed_doc = self.parseString(docstring)

            description = parsed_doc.description
            for attr in parsed_doc.attrs:
                name = attr.getName().strip()
                if name == "param":
                    params[attr.name] = attr.description.strip()
                elif name == "return":
                    returns = attr.description.strip()
                elif name == "raise":
                    raises[attr.error] = attr.description.strip()
                elif name == "consumes":
                    consumes.append(attr.description.strip())
                elif name == "produces":
                    produces.append(attr.description.strip())
                elif name == "version":
                    version = attr.description.strip()

        ret = JsonObject(
            **{
                "summary": short_description,
                "description": description,
                "params": params,
                "returns": returns,
                "consumes": consumes,
                "produces": produces,
                "raises": raises,
                "version": version
            }
        )
        return ret



# -*- coding: utf-8 -*-
"""
Module api_resource.py
---------------------
    The base Api resource implementation.
"""
from .flask import FlaskResource
from .swagger import SwaggerResource


class Resource(FlaskResource, SwaggerResource):
    """
    A Base API endpoint class implementation.
    """
    pass



# -*- coding: utf-8 -*-
"""
properties module: message properties/attributes definitions.
"""
from .base import BaseProperty
from .base import NormalProperty
from .base import ChoiceProperty
from .string_property import StringProperty
from .boolean_property import BooleanProperty
from .float_property import FloatProperty
from .integer_property import IntegerProperty
from .date_property import DateProperty
from .datetime_property import DateTimeProperty
from .object_property import ObjectProperty
from .array_property import ArrayProperty
from .dict_property import DictProperty
from .file_property import FileProperty



"""
float property module - defines float attributes.
"""
from simple_mappers import properties
from .base import BaseProperty
from werkzeug import exceptions

class FloatProperty(properties.FloatProperty, BaseProperty):
    """
    Store a floating point value
    """

    def inflate(self, value):
        """Returns the value that should be used to fill the object being mapped."""
        try:
            value = super().inflate(value)
            return value
        except ValueError:
            name = self.mapping_from or self.property_name
            raise exceptions.BadRequest(
                "On property: '{}'.\n"
                "Value: '{}'.\n"
                "Error: Expected float type.".format(name, value)
            )

    def to_swagger(self, schema:dict):
        """
        Serializes attribute to swagger specification format.
        :param schema: a swagger spec dict.
        :return: the modified swagger spec.
        """
        super(FloatProperty, self).to_swagger(schema)
        schema["type"] = "float"
        return schema



# -*- coding: utf-8 -*-
"""
Module path_utils.py
---------------------
    utility functions to dealing with paths during serialization
"""
from mercury.parameters import path
import inspect
import re


# route substring replace map
ROUTE_REPLACE_MAP = {
    "int:": "",
    "float:": "",
    "path:": "",
    "any:": "",
    "uuid:": "",
    # Escape all the characters in pattern except ASCII letters, numbers and '_'.
    "\\<": "{",
    "\\>": "}",
    ":":  ""
}
# the route replace regex
ROUTE_REPLACE_REGEX = re.compile("|".join(ROUTE_REPLACE_MAP.keys()))


def format_base_path_swagger(cls):
    """
    Returns the swagger base path
    :param cls: a resource class.
    :return: the base resource path
    """
    # checks for route meta data
    if cls.ROUTE is not None:
        swagger_path = cls.ROUTE
        # then formats the parameters correctly
        swagger_path = ROUTE_REPLACE_REGEX.sub(
            lambda match: ROUTE_REPLACE_MAP[re.escape(match.group(0))],
            swagger_path
        )
        # swagger_path = swagger_path.replace(":", "")
    else:
        # otherwise uses the module dotted path as path
        mod = cls.__module__.split('.')

        if len(mod) > 0 and cls.__name__.lower() != mod[-1]:
            mod = "/".join(mod)
            swagger_path = "/{}/{}".format(mod, cls.__name__.lower())
        else:
            mod = "/".join(mod)
            swagger_path = "/{}".format(mod)

    return swagger_path


# ROUTE_REGEX = re.compile(
#     r"<int:\W+>|"
#     r"<float:\W+>|"
#     r"<path:\W+>|"
#     r"<any:\W+>|"
#     r"<uuid:\W+>|"
#     r"<\W+>",
#     re.IGNORECASE
# )
ROUTE_REGEX = re.compile(
    r"<([A-Za-z_0-9]*:?)[A-Za-z0-9_]*>",
    re.IGNORECASE
)

def get_base_path_parameters(cls):
    """
    Returns a dictionary containing the base path parameter names as key and
    it type's as value.
    :param cls: a resource class.
    :return: the base path parameters.
    """
    # prepare swagger paths
    params = dict()
    # checks for route meta data
    if cls.ROUTE:
        swagger_path = cls.ROUTE
        # searches for path parameters on the route
        path_parameters = ROUTE_REGEX.findall(swagger_path)
        # if found any query parameter on the route
        if len(path_parameters) > 0:
            for p in path_parameters:
                p = p.replace("<", "").replace(">", "").split(":")
                if len(p) > 1:
                    params[p[1]] = p[0]
                else:
                    params[p[0]] = None
    return params


def check_route_and_verb_path_parameter_specifications(cls, verb):
    """Checks the base route parameters and the verb function parameter definitions and
    raises an error when the verb does not defines some of the route defined parameters."""
    method = getattr(cls, verb, None)
    if method is not None:
        method_params = method.__params__

        # sanity check in the base_path parameters and method parameters
        p_name = [p.name for p in method_params]
        for param_name in get_base_path_parameters(cls).keys():
            if param_name not in p_name:
                raise NotImplementedError(
                    "The path parameter '{parameter}' defined on the base_path of the class '{cls}'"
                    "is not defined on the method '{method}' for the verb '{verb}'."
                    "The method '{method}' must receive all path_parameters as its parameters."
                    "".format(parameter=param_name, cls=cls.__qualname__, method=method.__qualname__, verb=verb)
                )


def find_additional_params(cls, verb):
    """
    Returns a list of parameters that are not defined on the base path.
    :param cls: a resource class.
    :param verb: the verb implementation.
    :return: a list of parameters that are not defined on the base path.
    """
    base_params = get_base_path_parameters(cls).keys()
    ret = list()
    for par in verb.__params__:
        if issubclass(par.annotation, path.BasePathParameter) and par.name not in base_params:
            ret.append(par)

    return ret


def filter_required_params(param_list):
    """
    Removes all parameters that has default value defined from the param_list and returns it.
    :param param_list: a list of verb parameters
    :return: a list of required parameters
    """
    ret = list()
    for par in param_list:
        if issubclass(par.annotation, path.BasePathParameter):
            if par.default is inspect._empty:
                ret.append('{'+par.name+'}')
    return ret


def filter_not_required_params(param_list):
    """
    Removes all parameters that has default value defined from the param_list and returns it.
    :param param_list: a list of verb parameters
    :return: a list of required parameters
    """
    ret = list()
    for par in param_list:
        if issubclass(par.annotation, path.BasePathParameter):
            if par.default is not inspect._empty:
                ret.append('{'+par.name+'}')
    return ret



# -*- coding: utf-8 -*-
"""
Module string.py:
-----------------------
 Flask-Mercury Boolean query parameter type.
"""
from .base import BaseQueryParameter
from werkzeug.exceptions import BadRequest
import inspect


class Boolean(BaseQueryParameter):
    """
    Boolean query parameter type.
    """
    _booleans = ['true', 'false']

    def __new__(cls, name, value):
        """
        Boolean query parameter constructor.
        :param value: a value to be casted to str.
        :return: the value casted to str type
        """
        try:
            value = str(value)
            return value.lower() == 'true'
        except ValueError as e:
            raise BadRequest(
                "Wrong type for the query parameter named '{}'. "
                "It should be a boolean but we got '{}'.".format(name, value)
            )

    @classmethod
    def to_swagger(cls, param, doc):
        """
        Swagger spec query parameter serialization function.
        :param param: function parameter pointer.
        :param doc: the parameter doc_string.
        :return: the parameter spec.
        """
        spec = super().to_swagger(param, doc)
        spec['type'] = 'boolean'

        return spec

    @classmethod
    def __instancecheck__(cls, instance):
        """
        The magic function called to implement isinstance(..) functionality.
        :param instance: the object being evaluated.
        :return: true if instance should be considered a (direct or indirect) instance of class.
        """
        is_bool = str(instance).lower() in cls._booleans
        # builtin int type should be considered a instance of the Integer cls.
        if is_bool:
            return True

        if type(instance) is cls:
            return True

        # TODO add parameter name to the message
        raise BadRequest(
            "Wrong type for the query parameter. "
            "It should be boolean but we got '{}'.".format(instance)
        )


# -*- coding: utf-8 -*-
"""
Package resource
---------------------
Api resource provides api resource definition and implementation
for flask-mercury.
"""

_verbs = ["get", "post", "put", "delete", "head", "patch"]

from .meta_resource import BaseResource
from .api_resource import Resource



# -*- coding: utf-8 -*-
"""
Datetime property module - defines datetime attributes.
"""
from simple_mappers import properties
from .base import BaseProperty
from werkzeug import exceptions
import datetime

class DateTimeProperty(properties.DateTimeProperty, BaseProperty):

    def __init__(self, mask=None, **kwargs):
        """
        Datetime property type initialization.
        :param mask: string datetime mask used to parse string dates.
                    or a dict mapping the possible incoming values to its transformed values.

        :param kwargs required: True if the attribute is required.
                         When true and it is not possible to map the attribute from the other object,
                         it raises a value error exception. 
        :param kwargs default: default value or function generator.
        :param kwargs source_name: the name of the attribute in the other object
        """
        if mask is not None:
            self.mask = mask
        else:
            self.mask = "%d/%m/%Y %H:%M"
        
        super(DateTimeProperty, self).__init__(mask, **kwargs)
        
        # default value must be set as string
        if isinstance(self.default,(datetime.datetime, datetime.date,)):
            self.default = self.deflate(self.default)

    def inflate(self, value):
        """Returns the value that should be used to fill the object being mapped."""
        try:
            value = super().inflate(value)
            return value
        except TypeError as e:
            raise exceptions.BadRequest(str(e))
        except ValueError as e:
            name = self.mapping_from or self.property_name
            raise exceptions.BadRequest(
                "On Property: {}\n"
                "Value: {}\n"
                "Error: Date property '{}' does not match format '{}'.".format(name, value, name, self.mask)
            )

    def to_swagger(self, schema:dict):
        """
        Serializes attribute to swagger specification format.
        :param schema: a swagger spec dict.
        :return: the modified swagger spec.
        """
        super(DateTimeProperty,self).to_swagger(schema)
        schema["type"] = "string"
        schema["format"] = self.mask
        return schema


# -*- coding: utf-8 -*-
"""
Module integer_tests.py
--------------------------
A test suite for the Uuid query parameter.
"""
import unittest
from mercury.parameters.query.uuid import Uuid
from mercury.doc_parser import DocString
import inspect
from werkzeug.exceptions import BadRequest
import uuid


def test_function(parameter:Uuid="true"):
    """
    A function that defines a base parameter with default value.
    :param parameter: the function parameter with default value.
    """
    return True


class UuidParameterTests(unittest.TestCase):
    """
    A test suite for the String parameter class.
    """

    def test_should_serialize_spec_parameter_correctly(self):
        """
        Tests if the parameter spec is serialized correctly.
        """

        signature = inspect.signature(test_function)
        method_params = list(signature.parameters.values())
        doc_parser = DocString()
        doc_str = doc_parser.parse(test_function.__doc__)
        par = method_params[0]
        spec = Uuid.to_swagger(par, doc_str.to_dict().get("params").get("parameter"))

        self.assertEqual(spec['type'], "string")

    def test_should_initialize_str(self):
        """
        test boolean parameter initialization function.
        """
        val_str = str(uuid.uuid4())
        val = Uuid("parameter", val_str)
        self.assertTrue(isinstance(val, uuid.UUID))
        self.assertTrue(val == uuid.UUID(val_str))

    def test_should_pass_isinstance(self):
        """
        test if a bool builtin type pass in the isinstance(..) check.
        """
        val = uuid.uuid4()

        self.assertTrue(isinstance(val, Uuid))
        self.assertTrue(isinstance(str(val), Uuid))
        # flag = "11.1"
        # self.assertTrue(isinstance(flag, Float))

    def test_should_not_pass_instance_check(self):
        flag = 11
        self.assertRaises(BadRequest, lambda: isinstance(flag, Uuid))

    def test_should_call_test_function(self):
        val = uuid.uuid4()
        ret = test_function(str(val))
        self.assertTrue(ret)


if __name__ == "__main__":
    unittest.main()


# -*- coding: utf-8 -*-
"""
Module mercury_app.py
---------------------
  MercuryApp provides swagger endpoints and initialization functions to the Flask-Mercury framework.
"""
import json
import flask
from simple_mappers import JsonObject
import inspect
import importlib
import pkgutil
from .resource import Resource
from .api_model import ApiModel
from flask import Flask
from .doc_parser import DocString


class MercuryApp(object):
    """
    Swagger ui endpoints.
    """
    def __init__(self, app: Flask, api_title, schemes, api_doc):
        self.app = app
        # swagger specification object initialization
        swagger_spec = dict(
            definitions=dict(),
            # initialize the paths part
            paths=dict(),
            tags=list(),
            swagger="2.0",
            info=JsonObject(),
            basePath='/' + api_title,
            schemes=schemes
        )
        self.spec = JsonObject(**swagger_spec)
        self.spec.info.description = api_doc.summary
        self.spec.info.version = api_doc.version
        # self.spec.info.version = "1.0.0"
        self.spec.info.title = api_title
        # TODO
        # swagger_spec.info.termsOfService = ""
        # swagger_spec.info.contact = {"email":""}
        # swagger_spec.info.license = {"name": doc.license, "url":""}

        self.index_path = "index.html"
        self.api_title = api_title

    def auto_discover_spec(self, api_module):
        def selector(member):
            """Filters members that are either an ApiModel implementation or a Resource implementation"""
            if inspect.isclass(member):
                if issubclass(member, ApiModel) and member is not ApiModel:
                    return True
                elif issubclass(member, Resource) and member is not Resource:
                    return True
            return False

        # searches for 'mercury.Resource' implementation recursively on the API_ROOT package
        if hasattr(api_module, "__path__"):
            path = api_module.__path__
            prefix = api_module.__name__ + "."
        else:  # when api module is the __main__ module it has not a __path__ attr
            path = api_module.__file__
            prefix = ""

        for finder, name, is_pack in pkgutil.walk_packages(path, prefix):
            if is_pack:
                # skip packages
                continue
            # import the module if it has not been imported yet
            print(">>>>", name)
            mod = importlib.import_module(name)
            # filter mercury members
            mod_members = inspect.getmembers(
                mod,
                predicate=selector
            )
            # for each member
            for member_name, member_obj in mod_members:
                if issubclass(member_obj, Resource):
                    spec = member_obj.to_swagger()
                    self.add_resource_spec(spec)
                    member_obj.register(self.app, self.api_title)
                elif issubclass(member_obj, ApiModel):
                    member_obj.to_swagger(self.spec)

    def add_resource_spec(self, resource_spec):
        for path, path_spec in resource_spec.paths.items():
            if path in self.spec.paths:
                for verb, verb_spec in path_spec.items():
                    if verb not in self.spec.paths[path]:
                        self.spec.paths[path][verb] = verb_spec
            else:
                self.spec.paths[path] = dict()
                for verb, verb_spec in path_spec.items():
                    self.spec.paths[path][verb] = verb_spec

        for tag in resource_spec.tags:
            if tag not in self.spec.tags:
                self.spec.tags.append(tag)

    def swagger_json(self):
        """
        endpoint that returns the swagger.json specification.
        """
        return json.dumps(self.spec.to_dict())

    def index(self):
        """
        returns the swagger index.html page.
        """
        spec_name = "{}.swagger_json".format(self.api_title)
        temp = flask.render_template(self.index_path, spec=spec_name, api_title=self.api_title)
        return temp


def load(app:Flask, api_title, schemes, api_package="."):
    """ Flask-Mercury loader function.
    An autodiscover function used to load the framework on the current project.
    It will scan the current project looking for mercury Resources and Models
    :param app: The flask app reference.
    :param api_title: the api title.
    :param schemes: a list containing two possible values [http, https].
    :param api_package: a dotted path to the api root package.
    """
    # some imports needed to load the module
    # we are importing them here because it is required only on the first run
    import importlib
    from . import utils

    # builds api info part of the specification
    api_module = importlib.import_module(api_package, app.import_name)
    api_doc = api_module.__doc__
    doc_parser = DocString()
    doc = doc_parser.parse(api_doc)
    # doc = utils.parse_docstring(api_doc)
    s = MercuryApp(app, api_title, schemes, doc)
    s.auto_discover_spec(api_module)

    # finally register swagger routes
    # Setting CORS due to swagger-ui
    from flask_cors import CORS
    from flask import Blueprint
    import os
    current_dir = os.path.abspath(os.path.dirname(__file__))

    # Blueprint initialization
    blueprint = Blueprint(
        api_title,
        __name__,
        static_folder="{}/static".format(current_dir),
        template_folder="{}/static/templates".format(current_dir),
        url_prefix=api_title
    )
    CORS(blueprint)

    blueprint.add_url_rule('/swagger', view_func=s.index)
    blueprint.add_url_rule('/swagger/swagger.json', endpoint="swagger_json", view_func=s.swagger_json)
    app.register_blueprint(blueprint, url_prefix="/"+blueprint.name, static_folder="{}/static".format(current_dir))



# -*- coding: utf-8 -*-
"""
Module path_utils.py
---------------------
    utility functions to dealing with paths during serialization
"""
from mercury.parameters import path
import inspect
import re


# ROUTE_REGEX = re.compile(r"^/\W+/$", re.IGNORECASE)
ROUTE_REGEX = re.compile(
    r"^<int:\W+>|"
    r"<float:\W+>|"
    r"<path:\W+>|"
    r"<any:\W+>|"
    r"<uuid:\W+>|"
    r"<\W+>$",
    re.IGNORECASE
)


def get_route_parameters(route_path):
    """
    Returns a list of parameters defined on the flask route string.
    :param route_path: a route path string.
    :return: a list of parameter names
    """
    # get the resource base path
    # checks for route meta data
    parameters = list()
    # builds a base path parameter lists
    for p in ROUTE_REGEX.findall(route_path):
        p.replace("<", "").split(":")
        if len(p) > 1:
            parameters.append(p[1])
        else:
            parameters.append(p[0])
    return


"""
Dict property module - defines dictionary attributes.
"""
from simple_mappers import properties
from .base import BaseProperty
from werkzeug import exceptions


class DictProperty(properties.DictProperty, BaseProperty):
    """
    Stores a dict of items
    """

    def inflate(self, value):
        """Returns the value that should be used to fill the object being mapped."""
        try:
            value = super().inflate(value)
            return value
        except TypeError as e:
            raise exceptions.BadRequest(str(e))

    def to_swagger(self, schema: dict):
        """
        Serializes attribute to swagger specification format.
        :param schema: a swagger spec dict.
        :return: the modified swagger spec.
        """
        super(DictProperty, self).to_swagger(schema)
        schema["type"] = "object"
        return schema



# -*- coding: utf-8 -*-
"""
Module string_property.py
----------------------------
 Defines string attributes.
"""
from simple_mappers import properties
from .base import BaseProperty
from werkzeug import exceptions


class StringProperty(properties.StringProperty, BaseProperty):
    """
    Defines an string property mapper object.
    """

    def inflate(self, value):
        """Returns the value that should be used to fill the object being mapped."""
        try:
            if self.required and not isinstance(value, str):
                name = self.mapping_from or self.property_name
                raise exceptions.BadRequest(
                    "On property: '{}'.\n"
                    "Value: '{}'.\n"
                    "Error: Expected string type.".format(name, value)
                )

            value = super().inflate(value)
            return value
        except ValueError:
            name = self.mapping_from or self.property_name
            raise exceptions.BadRequest(
                "On property: '{}'.\n"
                "Value: '{}'.\n"
                "Error: Expected string type.".format(name, value)
            )

    def to_swagger(self, schema: dict):
        """
        Serializes attribute to swagger specification format.
        :param schema: a swagger spec dict.
        :return: the modified swagger spec.
        """
        super(StringProperty,self).to_swagger(schema)
        schema["type"] = "string"
        return schema


# -*- coding: utf-8 -*-
"""
Module base_tests.py
----------------------
A test suite for the float path parameter.
"""
import unittest
from mercury.parameters.path.float_param import Float
from mercury.doc_parser import DocString
import inspect
from werkzeug.exceptions import BadRequest


def test_function(parameter:Float="true"):
    """
    A function that defines a base parameter with default value.
    :param parameter: the function parameter with default value.
    """
    return True


class FloatParameterTests(unittest.TestCase):
    """
    A test suite for the base parameter class.
    """

    def test_should_serialize_spec_parameter_correctly(self):
        """
        Tests if the parameter spec is serialized correctly.
        """

        signature = inspect.signature(test_function)
        method_params = list(signature.parameters.values())
        doc_parser = DocString()
        doc_str = doc_parser.parse(test_function.__doc__)
        par = method_params[0]
        spec = Float.to_swagger(par, doc_str.to_dict().get("params").get("parameter"))

        self.assertEqual(spec['type'], "number")

    def test_should_initialize_float(self):
        """
        test boolean parameter initialization function.
        """
        val = Float("parameter", "1.1")
        self.assertTrue(isinstance(val, float))
        self.assertTrue(val == 1.1)

        val = Float("parameter", 1.1)
        self.assertTrue(isinstance(val, float))
        self.assertTrue(val == 1.1)

    def test_should_not_initialize_float(self):
        """
        test boolean parameter initialization function.
        """
        self.assertRaises(BadRequest, lambda: Float("parameter", "false"))

    def test_should_pass_isinstance(self):
        """
        test if a bool builtin type pass in the isinstance(..) check.
        """
        flag = 11.1
        self.assertTrue(isinstance(flag, Float))
        # flag = "11.1"
        # self.assertTrue(isinstance(flag, Float))

    def test_should_not_pass_instance_check(self):
        flag = "sss"
        self.assertRaises(BadRequest, lambda: isinstance(flag, Float))

    def test_should_call_test_function(self):
        ret = test_function(11.1)
        self.assertTrue(ret)
        ret = test_function(11)
        self.assertTrue(ret)


if __name__ == "__main__":
    unittest.main()


# -*- coding: utf-8 -*-
"""
Module setup.py
-----------------
 mercury setuptool module.
"""
from setuptools import setup, find_packages
# builds the project dependency list
install_requires = None
with open('requirements.txt', 'r') as f:
        install_requires = f.readlines()

# setup function call
setup(
    name="flask-mercury",
    version="1.1.8",
    author="Luis Felipe Muller",
    author_email="luisfmuller@gmail.com",
    description=(
        "Flask-Mercury is a easy to used framework for "
        "implementing and documenting restful APIs based on flask and swagger."
    ),
    keywords="Flask, Swagger, APIs, Rest",
    # Install project dependencies
    install_requires=install_requires,
    package_data={
        # If any package contains *.txt or *.rst files, include them:
        '': ['*.txt', '*.rst', '*.md', "*.json", "*.zip", "*.js", "*.css", "*.html", "*.png"],
    },
    include_package_data=True,
    packages=find_packages(exclude=["*tests"])
)



# -*- coding: utf-8 -*-
"""
Flask-Mercury
-----------------
Flask-Mercury is a simple lightweight framework to build flask apis
and swagger documentation.
It is empowered by simple_mappers to delivers auto-mapping capabilities.
"""
from .resource import Resource
from .api_model import ApiModel
from . import parameters
from .mercury_app import load




# -*- coding: utf-8 -*-
"""
Module exceptions.py
-----------------------------------
Mercury error exceptions classes.
"""


class ResourceSpecError(Exception):
    """
    Exception class representing errors of a mercury resource specification/implementation.
    """
    pass



# -*- coding: utf-8 -*-
"""
Module base_tests.py
----------------------
A test suite for the base path parameter.
"""
import unittest
from mercury.parameters.path.boolean import Boolean
from mercury.doc_parser import DocString
import inspect
from werkzeug.exceptions import BadRequest


def test_function(parameter:Boolean="true"):
    """
    A function that defines a base parameter with default value.
    :param parameter: the function parameter with default value.
    """
    return True


class BooleanParameterTests(unittest.TestCase):
    """
    A test suite for the base parameter class.
    """

    def test_should_serialize_spec_parameter_correctly(self):
        """
        Tests if the parameter spec is serialized correctly.
        """

        signature = inspect.signature(test_function)
        method_params = list(signature.parameters.values())
        doc_parser = DocString()
        doc_str = doc_parser.parse(test_function.__doc__)
        par = method_params[0]
        spec = Boolean.to_swagger(par, doc_str.to_dict().get("params").get("parameter"))

        self.assertEqual(spec['type'], "boolean")

    def test_should_initialize_boolean_as_true(self):
        """
        test boolean parameter initialization function.
        """
        val = Boolean("parameter", "true")
        self.assertTrue(isinstance(val, bool))
        self.assertTrue(val)

        val = Boolean("parameter", True)
        self.assertTrue(isinstance(val, bool))
        self.assertTrue(val)

    def test_should_initialize_boolean_as_false(self):
        """
        test boolean parameter initialization function.
        """
        val = Boolean("parameter", "false")
        self.assertTrue(isinstance(val, bool))
        self.assertTrue(not val)

        val = Boolean("parameter", False)
        self.assertTrue(isinstance(val, bool))
        self.assertTrue(not val)

    def test_should_pass_isinstance(self):
        """
        test if a bool builtin type pass in the isinstance(..) check.
        """
        flag = True

        self.assertTrue(isinstance(flag, Boolean))
        flag = False

        self.assertTrue(isinstance(flag, Boolean))

        flag = "true"

        self.assertTrue(isinstance(flag, Boolean))

        flag = "false"

        self.assertTrue(isinstance(flag, Boolean))

    def test_should_not_pass_instance_check(self):
        try:
            flag = 21
            isinstance(flag, Boolean)
        except BadRequest as ex:
            self.assertTrue(isinstance(ex, BadRequest))

        try:
            flag = "not a flag"
            isinstance(flag, Boolean)
        except BadRequest as ex:
            self.assertTrue(isinstance(ex, BadRequest))

    def test_should_call_test_function(self):
        ret = test_function("true")
        self.assertTrue(ret)
        ret = test_function(True)
        self.assertTrue(ret)
        ret = test_function("False")
        self.assertTrue(ret)
        ret = test_function(False)
        self.assertTrue(ret)


if __name__ == "__main__":
    unittest.main()


# -*- coding: utf-8 -*-
"""
Package parameters:
-----------------------
 Flask-Mercury path parameter type definitions.
"""
from .base import BasePathParameter
from .boolean import Boolean
from .float_param import Float
from .integer import Integer
from .string import String, Any
from .uuid import Uuid



# -*- coding: utf-8 -*-
"""
boolean property module - defines boolean attributes.
"""
from simple_mappers import properties
from .base import BaseProperty


class BooleanProperty(properties.BooleanProperty, BaseProperty):
    """
    Stores a boolean value
    """
    def to_swagger(self, schema:dict):
        """
        Serializes attribute to swagger specification format.
        :param schema: a swagger spec dict.
        :return: the modified swagger spec.
        """
        super(BooleanProperty, self).to_swagger(schema)
        schema["type"] = "boolean"
        return schema



# -*- coding: utf-8 -*-
"""
Module meta_resource.py
-------------------------
    Metaclass implementation for the api resource
"""
from flask.views import MethodView
import inspect
from mercury import utils
from flask.views import MethodViewType
from . import _verbs
from ..doc_parser import DocString


class MetaResource(MethodViewType):
    """
    A Api Resource Meta class.
    """
    def __new__(mcs, what, bases=None, namespace=None):
        """
        Meta Resource constructor function.
        :param what: The resource class name.
        :param bases: A collection of superclasses.
        :param namespace: The namespace __dict__ for the class.
        """
        cls = super(MetaResource, mcs).__new__(mcs, what, bases, namespace)
        for verb in _verbs:
            mcs.get_signature(cls, verb)
        return cls

    def get_signature(cls, name):
        """
        Sets __params__, __pdoc__, __returns__ attribute to functions that are
        verb implementation.
        :param name: the function name.
        """
        if hasattr(cls, name):
            method = getattr(cls, name)
            signature = inspect.signature(method)
            method_params = list(signature.parameters.values())
            # add the meta attribute __params__ to the method dict
            object.__setattr__(method, "__params__", method_params)
            # add the meta attribute __pdoc__ containing the parsed method doc string
            docstring_parser = DocString()
            object.__setattr__(method, "__pdoc__", docstring_parser.parse(method.__doc__))
            # adds the meta attribute __returns__
            object.__setattr__(method, "__returns__", signature.return_annotation)


# the base resource class
BaseResource = MetaResource("BaseResource", (MethodView,), dict())



"""
Object property module - defines Object attributes.
"""
from simple_mappers import properties
from .base import BaseProperty


class ObjectProperty(properties.ObjectProperty, BaseProperty):
    """
    Stores a object property.
    """

    def to_swagger(self, schema:dict):
        """
        Serializes attribute to swagger specification format.
        :param schema: a swagger spec dict.
        :return: the modified swagger spec.
        """
        super(ObjectProperty, self).to_swagger(schema)
        if self.obj_type is not None:
            schema["$ref"] = "#/definitions/{}".format(self.obj_type.__name__)
        else:
            schema["type"] = "object"
            # schema["properties"] = {}
        return schema
